@@ -2,49 +2,41 @@ from __future__ import annotations
 
 import inspect
 import logging
+import os
 import sys
 import threading
+from concurrent.futures import ThreadPoolExecutor
 from contextlib import contextmanager
+from errno import ENOSYS
 from inspect import getframeinfo, stack
 from pathlib import Path, PurePath
-from stat import S_IWGRP, S_IWOTH, S_IWUSR
+from stat import S_IWGRP, S_IWOTH, S_IWUSR, filemode
 from types import TracebackType
-from typing import Callable, Iterator, Tuple, Type, Union
+from typing import TYPE_CHECKING, Any, Callable, Iterator, Tuple, Type, Union
+from uuid import uuid4
 
 import pytest
-from _pytest.logging import LogCaptureFixture
 
-from filelock import (
-    BaseFileLock,
-    FileLock,
-    SoftFileLock,
-    Timeout,
-    UnixFileLock,
-    WindowsFileLock,
-)
+from filelock import BaseFileLock, FileLock, SoftFileLock, Timeout, UnixFileLock, WindowsFileLock
+
+if TYPE_CHECKING:
+    from pytest_mock import MockerFixture
 
 
-@pytest.mark.parametrize(
-    ("lock_type", "path_type"),
-    [
-        (FileLock, str),
-        (FileLock, PurePath),
-        (FileLock, Path),
-        (SoftFileLock, str),
-        (SoftFileLock, PurePath),
-        (SoftFileLock, Path),
-    ],
-)
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+@pytest.mark.parametrize("path_type", [str, PurePath, Path])
+@pytest.mark.parametrize("filename", ["a", "new/b", "new2/new3/c"])
 def test_simple(
     lock_type: type[BaseFileLock],
-    path_type: type[str] | type[Path],
+    path_type: type[str | Path],
+    filename: str,
     tmp_path: Path,
-    caplog: LogCaptureFixture,
+    caplog: pytest.LogCaptureFixture,
 ) -> None:
     caplog.set_level(logging.DEBUG)
 
     # test lock creation by passing a `str`
-    lock_path = tmp_path / "a"
+    lock_path = tmp_path / filename
     lock = lock_type(path_type(lock_path))
     with lock as locked:
         assert lock.is_locked
@@ -78,6 +70,10 @@ def tmp_path_ro(tmp_path: Path) -> Iterator[Path]:
 
 @pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
 @pytest.mark.skipif(sys.platform == "win32", reason="Windows does not have read only folders")
+@pytest.mark.skipif(
+    sys.platform != "win32" and os.geteuid() == 0,
+    reason="Cannot make a read only file (that the current user: root can't read)",
+)
 def test_ro_folder(lock_type: type[BaseFileLock], tmp_path_ro: Path) -> None:
     lock = lock_type(str(tmp_path_ro / "a"))
     with pytest.raises(PermissionError, match="Permission denied"):
@@ -93,18 +89,50 @@ def tmp_file_ro(tmp_path: Path) -> Iterator[Path]:
 
 
 @pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+@pytest.mark.skipif(
+    sys.platform != "win32" and os.geteuid() == 0,
+    reason="Cannot make a read only file (that the current user: root can't read)",
+)
 def test_ro_file(lock_type: type[BaseFileLock], tmp_file_ro: Path) -> None:
     lock = lock_type(str(tmp_file_ro))
     with pytest.raises(PermissionError, match="Permission denied"):
         lock.acquire()
 
 
-@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
-def test_missing_directory(lock_type: type[BaseFileLock], tmp_path_ro: Path) -> None:
-    lock_path = tmp_path_ro / "a" / "b"
-    lock = lock_type(str(lock_path))
+WindowsOnly = pytest.mark.skipif(sys.platform != "win32", reason="Windows only")
 
-    with pytest.raises(OSError, match="No such file or directory:"):
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+@pytest.mark.parametrize(
+    ("expected_error", "match", "bad_lock_file"),
+    [
+        pytest.param(FileNotFoundError, "No such file or directory:", "", id="blank_filename"),
+        pytest.param(ValueError, "embedded null (byte|character)", "\0", id="null_byte"),
+        # Should be PermissionError on Windows
+        pytest.param(PermissionError, "Permission denied:", ".", id="current_directory")
+        if sys.platform == "win32"
+        else (
+            # Should be IsADirectoryError on MacOS and Linux
+            pytest.param(IsADirectoryError, "Is a directory", ".", id="current_directory")
+            if sys.platform in ["darwin", "linux"]
+            else
+            # Should be some type of OSError at least on other operating systems
+            pytest.param(OSError, None, ".", id="current_directory")
+        ),
+    ]
+    + [pytest.param(OSError, "Invalid argument", i, id=f"invalid_{i}", marks=WindowsOnly) for i in '<>:"|?*\a']
+    + [pytest.param(PermissionError, "Permission denied:", i, id=f"permission_{i}", marks=WindowsOnly) for i in "/\\"],
+)
+@pytest.mark.timeout(5)  # timeout in case of infinite loop
+def test_bad_lock_file(
+    lock_type: type[BaseFileLock],
+    expected_error: type[Exception],
+    match: str,
+    bad_lock_file: str,
+) -> None:
+    lock = lock_type(bad_lock_file)
+
+    with pytest.raises(expected_error, match=match):
         lock.acquire()
 
 
@@ -182,13 +210,12 @@ class ExThread(threading.Thread):
     def run(self) -> None:
         try:
             super().run()
-        except Exception:  # pragma: no cover
+        except Exception:  # noqa: BLE001 # pragma: no cover
             self.ex = sys.exc_info()  # pragma: no cover
 
     def join(self, timeout: float | None = None) -> None:
         super().join(timeout=timeout)
         if self.ex is not None:
-            print(f"fail from thread {self.name}")  # pragma: no cover
             raise RuntimeError from self.ex[1]  # pragma: no cover
 
 
@@ -334,8 +361,8 @@ def test_context_release_on_exc(lock_type: type[BaseFileLock], tmp_path: Path) -
         with lock as lock_1:
             assert lock is lock_1
             assert lock.is_locked
-            raise Exception
-    except Exception:
+            raise ValueError  # noqa: TRY301
+    except ValueError:
         assert not lock.is_locked
 
 
@@ -349,8 +376,8 @@ def test_acquire_release_on_exc(lock_type: type[BaseFileLock], tmp_path: Path) -
         with lock.acquire() as lock_1:
             assert lock is lock_1
             assert lock.is_locked
-            raise Exception
-    except Exception:
+            raise ValueError  # noqa: TRY301
+    except ValueError:
         assert not lock.is_locked
 
 
@@ -382,9 +409,8 @@ def test_del(lock_type: type[BaseFileLock], tmp_path: Path) -> None:
 def test_cleanup_soft_lock(tmp_path: Path) -> None:
     # tests if the lock file is removed after use
     lock_path = tmp_path / "a"
-    lock = SoftFileLock(str(lock_path))
 
-    with lock:
+    with SoftFileLock(lock_path):
         assert lock_path.exists()
     assert not lock_path.exists()
 
@@ -396,9 +422,9 @@ def test_poll_intervall_deprecated(lock_type: type[BaseFileLock], tmp_path: Path
 
     with pytest.deprecated_call(match="use poll_interval instead of poll_intervall") as checker:
         lock.acquire(poll_intervall=0.05)  # the deprecation warning will be captured by the checker
-        frameinfo = getframeinfo(stack()[0][0])  # get frameinfo of current file and lineno (+1 than the above lineno)
+        frame_info = getframeinfo(stack()[0][0])  # get frame info of current file and lineno (+1 than the above lineno)
         for warning in checker:
-            if warning.filename == frameinfo.filename and warning.lineno + 1 == frameinfo.lineno:  # pragma: no cover
+            if warning.filename == frame_info.filename and warning.lineno + 1 == frame_info.lineno:  # pragma: no cover
                 break
         else:  # pragma: no cover
             pytest.fail("No warnings of stacklevel=2 matching.")
@@ -418,15 +444,245 @@ def test_context_decorator(lock_type: type[BaseFileLock], tmp_path: Path) -> Non
     assert not lock.is_locked
 
 
+def test_lock_mode(tmp_path: Path) -> None:
+    # test file lock permissions are independent of umask
+    lock_path = tmp_path / "a.lock"
+    lock = FileLock(str(lock_path), mode=0o666)
+
+    # set umask so permissions can be anticipated
+    initial_umask = os.umask(0o022)
+    try:
+        lock.acquire()
+        assert lock.is_locked
+
+        mode = filemode(lock_path.stat().st_mode)
+        assert mode == "-rw-rw-rw-"
+    finally:
+        os.umask(initial_umask)
+
+    lock.release()
+
+
+def test_lock_mode_soft(tmp_path: Path) -> None:
+    # test soft lock permissions are dependent of umask
+    lock_path = tmp_path / "a.lock"
+    lock = SoftFileLock(str(lock_path), mode=0o666)
+
+    # set umask so permissions can be anticipated
+    initial_umask = os.umask(0o022)
+    try:
+        lock.acquire()
+        assert lock.is_locked
+
+        mode = filemode(lock_path.stat().st_mode)
+        if sys.platform == "win32":
+            assert mode == "-rw-rw-rw-"
+        else:
+            assert mode == "-rw-r--r--"
+    finally:
+        os.umask(initial_umask)
+
+    lock.release()
+
+
+def test_umask(tmp_path: Path) -> None:
+    lock_path = tmp_path / "a.lock"
+    lock = FileLock(str(lock_path), mode=0o666)
+
+    initial_umask = os.umask(0)
+    os.umask(initial_umask)
+
+    lock.acquire()
+    assert lock.is_locked
+
+    current_umask = os.umask(0)
+    os.umask(current_umask)
+    assert initial_umask == current_umask
+
+    lock.release()
+
+
+def test_umask_soft(tmp_path: Path) -> None:
+    lock_path = tmp_path / "a.lock"
+    lock = SoftFileLock(str(lock_path), mode=0o666)
+
+    initial_umask = os.umask(0)
+    os.umask(initial_umask)
+
+    lock.acquire()
+    assert lock.is_locked
+
+    current_umask = os.umask(0)
+    os.umask(current_umask)
+    assert initial_umask == current_umask
+
+    lock.release()
+
+
 def test_wrong_platform(tmp_path: Path) -> None:
     assert not inspect.isabstract(UnixFileLock)
     assert not inspect.isabstract(WindowsFileLock)
     assert inspect.isabstract(BaseFileLock)
 
     lock_type = UnixFileLock if sys.platform == "win32" else WindowsFileLock
-    lock = lock_type(str(tmp_path / "lockfile"))
+    lock = lock_type(tmp_path / "lockfile")
 
     with pytest.raises(NotImplementedError):
         lock.acquire()
     with pytest.raises(NotImplementedError):
-        lock._release()
+        lock._release()  # noqa: SLF001
+
+
+@pytest.mark.skipif(sys.platform == "win32", reason="flock not run on windows")
+def test_flock_not_implemented_unix(tmp_path: Path, mocker: MockerFixture) -> None:
+    mocker.patch("fcntl.flock", side_effect=OSError(ENOSYS, "mock error"))
+    with pytest.raises(NotImplementedError), FileLock(tmp_path / "a.lock"):
+        pass
+
+
+def test_soft_errors(tmp_path: Path, mocker: MockerFixture) -> None:
+    mocker.patch("os.open", side_effect=OSError(ENOSYS, "mock error"))
+    with pytest.raises(OSError, match="mock error"):
+        SoftFileLock(tmp_path / "a.lock").acquire()
+
+
+def _check_file_read_write(txt_file: Path) -> None:
+    for _ in range(3):
+        uuid = str(uuid4())
+        txt_file.write_text(uuid)
+        assert txt_file.read_text() == uuid
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_thrashing_with_thread_pool_passing_lock_to_threads(tmp_path: Path, lock_type: type[BaseFileLock]) -> None:
+    def mess_with_file(lock_: BaseFileLock) -> None:
+        with lock_:
+            _check_file_read_write(txt_file)
+
+    lock_file, txt_file = tmp_path / "test.txt.lock", tmp_path / "test.txt"
+    lock = lock_type(lock_file)
+    with ThreadPoolExecutor() as executor:
+        results = [executor.submit(mess_with_file, lock) for _ in range(100)]
+    assert all(r.result() is None for r in results)
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_thrashing_with_thread_pool_global_lock(tmp_path: Path, lock_type: type[BaseFileLock]) -> None:
+    def mess_with_file() -> None:
+        with lock:
+            _check_file_read_write(txt_file)
+
+    lock_file, txt_file = tmp_path / "test.txt.lock", tmp_path / "test.txt"
+    lock = lock_type(lock_file)
+    with ThreadPoolExecutor() as executor:
+        results = [executor.submit(mess_with_file) for _ in range(100)]
+
+    assert all(r.result() is None for r in results)
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_thrashing_with_thread_pool_lock_recreated_in_each_thread(
+    tmp_path: Path,
+    lock_type: type[BaseFileLock],
+) -> None:
+    def mess_with_file() -> None:
+        with lock_type(lock_file):
+            _check_file_read_write(txt_file)
+
+    lock_file, txt_file = tmp_path / "test.txt.lock", tmp_path / "test.txt"
+    with ThreadPoolExecutor() as executor:
+        results = [executor.submit(mess_with_file) for _ in range(100)]
+
+    assert all(r.result() is None for r in results)
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_lock_can_be_non_thread_local(
+    tmp_path: Path,
+    lock_type: type[BaseFileLock],
+) -> None:
+    lock = lock_type(tmp_path / "test.lock", thread_local=False)
+
+    for _ in range(2):
+        thread = threading.Thread(target=lock.acquire, kwargs={"timeout": 2})
+        thread.start()
+        thread.join()
+
+    assert lock.lock_counter == 2
+
+    lock.release(force=True)
+
+
+def test_subclass_compatibility(tmp_path: Path) -> None:
+    class MyFileLock(FileLock):
+        def __init__(  # noqa: PLR0913 Too many arguments to function call (6 > 5)
+            self,
+            lock_file: str | os.PathLike[str],
+            timeout: float = -1,
+            mode: int = 0o644,
+            thread_local: bool = True,
+            my_param: int = 0,
+            **kwargs: dict[str, Any],
+        ) -> None:
+            pass
+
+    lock_path = tmp_path / "a"
+    MyFileLock(str(lock_path), my_param=1)
+
+    class MySoftFileLock(SoftFileLock):
+        def __init__(  # noqa: PLR0913 Too many arguments to function call (6 > 5)
+            self,
+            lock_file: str | os.PathLike[str],
+            timeout: float = -1,
+            mode: int = 0o644,
+            thread_local: bool = True,
+            my_param: int = 0,
+            **kwargs: dict[str, Any],
+        ) -> None:
+            pass
+
+    MySoftFileLock(str(lock_path), my_param=1)
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_singleton_and_non_singleton_locks_are_distinct(lock_type: type[BaseFileLock], tmp_path: Path) -> None:
+    lock_path = tmp_path / "a"
+    lock_1 = lock_type(str(lock_path), is_singleton=False)
+    assert lock_1.is_singleton is False
+
+    lock_2 = lock_type(str(lock_path), is_singleton=True)
+    assert lock_2.is_singleton is True
+    assert lock_2 is not lock_1
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_singleton_locks_are_the_same(lock_type: type[BaseFileLock], tmp_path: Path) -> None:
+    lock_path = tmp_path / "a"
+    lock_1 = lock_type(str(lock_path), is_singleton=True)
+
+    lock_2 = lock_type(str(lock_path), is_singleton=True)
+    assert lock_2 is lock_1
+
+
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_singleton_locks_are_distinct_per_lock_file(lock_type: type[BaseFileLock], tmp_path: Path) -> None:
+    lock_path_1 = tmp_path / "a"
+    lock_1 = lock_type(str(lock_path_1), is_singleton=True)
+
+    lock_path_2 = tmp_path / "b"
+    lock_2 = lock_type(str(lock_path_2), is_singleton=True)
+    assert lock_1 is not lock_2
+
+
+@pytest.mark.skipif(hasattr(sys, "pypy_version_info"), reason="del() does not trigger GC in PyPy")
+@pytest.mark.parametrize("lock_type", [FileLock, SoftFileLock])
+def test_singleton_locks_are_deleted_when_no_external_references_exist(
+    lock_type: type[BaseFileLock],
+    tmp_path: Path,
+) -> None:
+    lock_path = tmp_path / "a"
+    lock = lock_type(str(lock_path), is_singleton=True)
+
+    assert lock_type._instances == {str(lock_path): lock}  # noqa: SLF001
+    del lock
+    assert lock_type._instances == {}  # noqa: SLF001
